#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.tree import Block, Break, Heading, ListItem, Region, Rule, Text
import re

# Regular expressions.

syntax = {
    # Page regions:
    "regionstart"   : r"((^\s*)([{]{3,}))",                         # {{{...
    "regionend"     : r"^\s*([}]{3,})",                             # }}}...
    "header"        : r"#!(.*?)\n",                                 # #! char-excl-nl

    # Region contents:
    # Line-oriented patterns:
                      # blank line
    "break"         : r"^(\s*?)\n",
                      # [ws...] =... ws... expecting headingend
    "heading"       : r"^(\s*)(?P<x>=+)(\s+)(?=.*?\s+(?P=x)\s*\n)",
                      # indent list-item [ws...]
    "listitem"      : r"^(\s+)(\*)(\s*)",
                      # indent number-item ws...
    "listitem_num"  : r"^(\s+)(\d+\.)(\s+)",
                      # indent alpha-item ws...
    "listitem_alpha": r"^(\s+)([aA]\.)(\s+)",
                      # indent roman-item ws...
    "listitem_roman": r"^(\s+)([iI]\.)(\s+)",
                      # indent dot-item [ws...]
    "listitem_dot"  : r"^(\s+)(\.)(\s*)",

    # Region contents:
    # Inline patterns:
    "rule"          : r"(-----*)",                                  # ----...

    # Heading contents:
    "headingend"    : r"(\s+)(=+)(\s*\n)",                          # ws... =... [ws...] nl

    # List contents:
    "listitemend"   : r"^",                                         # next line
    }

# Define patterns for the regular expressions.

patterns = {}
for name, value in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | re.MULTILINE)



# Tokenising functions.

class TokenStream:

    "A stream of tokens taken from a string."

    def __init__(self, s):
        self.s = s
        self.pos = 0
        self.match = None
        self.matching = None

    def read_until(self, pattern_names, remaining=True):

        """
        Find the first match for the given 'pattern_names'. Return the text
        preceding any match, the remaining text if no match was found, or None
        if no match was found and 'remaining' is given as a false value.
        """

        first = None
        self.matching = None

        # Find the first matching pattern.

        for pattern_name in pattern_names:
            match = patterns[pattern_name].search(self.s, self.pos)
            if match:
                start, end = match.span()
                if self.matching is None or start < first:
                    first = start
                    self.matching = pattern_name
                    self.match = match

        if self.matching is None:
            if remaining:
                return self.s[self.pos:]
            else:
                return None
        else:
            return self.s[self.pos:first]

    def read_match(self, group=1):

        """
        Return the matched text, updating the position in the stream. If 'group'
        is specified, the indicated group in a match will be returned.
        Typically, group 1 should contain all pertinent data, but groups defined
        within group 1 can provide sections of the data.
        """

        if self.match:
            _start, self.pos = self.match.span()
            try:
                return self.match.group(group)
            except IndexError:
                return ""
        else:
            self.pos = len(self.s)
            return None



# Parser functions.

def parse_page(s):

    """
    Parse page text 's'. Pages consist of regions delimited by markers.
    """

    return parse_region(TokenStream(s))

def parse_region(items, level=0, indent=0):

    """
    Parse the data provided by 'items' to populate a region with the given
    'level' at the given 'indent'.
    """

    region = Region([], level, indent)

    # Parse section headers.

    parse_region_header(items, region)

    # Parse section body.

    if region.is_transparent():
        parse_region_wiki(items, region)
    else:
        parse_region_opaque(items, region)

    return region

def parse_region_header(items, region):

    """
    Parse the region header from the 'items', setting it for the given 'region'.
    """

    if items.read_until(["header"], False) == "": # None means no header
        region.type = items.read_match()

def parse_region_wiki(items, region):

    "Parse the data provided by 'items' to populate a wiki 'region'."

    new_block(region)
    parse_region_details(items, region, [
        "break", "heading", "listitem", "listitem_num", "listitem_alpha",
        "listitem_roman", "listitem_dot", "regionstart", "regionend", "rule"])

def parse_region_opaque(items, region):

    "Parse the data provided by 'items' to populate an opaque 'region'."

    parse_region_details(items, region, ["regionend"])

def parse_region_details(items, region, pattern_names):

    "Parse 'items' within 'region' searching using 'pattern_names'."

    try:
        while True:

            # Obtain text before any marker or the end of the input.

            preceding = items.read_until(pattern_names)
            if preceding:
                region.append_text(Text(preceding))

            # End of input.

            if not items.matching:
                break

            # Obtain any feature.

            feature = items.read_match()
            handler = handlers.get(items.matching)

            # Handle each feature or add text to the region.

            if handler:
                handler(items, region)
            else:
                region.append_text(Text(feature))

    except StopIteration:
        pass

    region.normalise()

def end_region(items, region):

    "End the parsing of 'region'."

    raise StopIteration

def parse_break(items, region):

    "Handle a paragraph break within 'region'."

    region.add(Break())
    new_block(region)

def parse_heading(items, region):

    "Handle a heading."

    start_extra = items.read_match(1)
    level = len(items.read_match(2))
    start_pad = items.read_match(3)
    heading = Heading([], level, start_extra, start_pad)
    parse_region_details(items, heading, ["headingend"])
    region.append(heading)
    new_block(region)

def parse_heading_end(items, heading):

    "Handle the end of a heading."

    level = len(items.read_match(2))
    if heading.level == level:
        heading.end_pad = items.read_match(1)
        heading.end_extra = items.read_match(3)
        raise StopIteration

def parse_listitem(items, region):

    "Handle a list item marker within 'region'."

    indent = len(items.read_match(1))
    marker = items.read_match(2)
    space = items.read_match(3)
    item = ListItem([], indent, marker, space)
    parse_region_details(items, item, ["listitemend"])
    region.append(item)
    new_block(region)

def parse_listitem_end(items, item):

    "Handle the end of a list."

    raise StopIteration

def parse_rule(items, region):

    "Handle a horizontal rule within 'region'."

    length = len(items.read_match(1))
    rule = Rule(length)
    region.append(rule)
    new_block(region)

def parse_section(items, region):

    "Handle the start of a new section within 'region'."

    # Parse the section and start a new block after the section.

    indent = len(items.read_match(2))
    level = len(items.read_match(3))
    region.append(parse_region(items, level, indent))
    new_block(region)

def parse_section_end(items, region):

    "Handle the end of a new section within 'region'."

    feature = items.read_match()
    if region.have_end(feature):
        raise StopIteration
    else:
        region.append_text(Text(feature))

# Pattern handlers.

handlers = {
    None : end_region,
    "break" : parse_break,
    "heading" : parse_heading,
    "headingend" : parse_heading_end,
    "listitemend" : parse_listitem_end,
    "listitem" : parse_listitem,
    "listitem_alpha" : parse_listitem,
    "listitem_dot" : parse_listitem,
    "listitem_num" : parse_listitem,
    "listitem_roman" : parse_listitem,
    "regionstart" : parse_section,
    "regionend" : parse_section_end,
    "rule" : parse_rule,
    }

def new_block(region):

    "Start a new block in 'region'."

    block = Block([])
    region.add(block)



# Top-level functions.

parse = parse_page

# vim: tabstop=4 expandtab shiftwidth=4
